
/* LoadBalanceType_publisher.c

A publication of data of type demo_rip_loadBalance_vector

This file is derived from code automatically generated by the rtiddsgen 
command:

rtiddsgen -language C -example <arch> LoadBalanceType.idl

Example publication of type demo_rip_loadBalance_vector automatically generated by 
'rtiddsgen'. To test them follow these steps:

(1) Compile this file and the example subscription.

(2) Start the subscription on the same domain used for RTI Data Distribution
Service with the command
objs/<arch>/LoadBalanceType_subscriber <domain_id> <sample_count>

(3) Start the publication on the same domain used for RTI Data Distribution
Service with the command
objs/<arch>/LoadBalanceType_publisher <domain_id> <sample_count>

(4) [Optional] Specify the list of discovery initial peers and 
multicast receive addresses via an environment variable or a file 
(in the current working directory) called NDDS_DISCOVERY_PEERS. 

You can run any number of publishers and subscribers programs, and can 
add and remove them dynamically from the domain.

Example:

To run the example application on domain <domain_id>:

On Unix: 

objs/<arch>/LoadBalanceType_publisher <domain_id> 
objs/<arch>/LoadBalanceType_subscriber <domain_id> 

On Windows:

objs\<arch>\LoadBalanceType_publisher <domain_id>  
objs\<arch>\LoadBalanceType_subscriber <domain_id>   
*/

// include for cuda
#include <cuda.h>

// include for dds
#include <stdio.h>
#include <stdlib.h>
#include "ndds/ndds_c.h"
#include "LoadBalanceType.h"
#include "LoadBalanceTypeSupport.h"

// CUDA Function for SPMV
void randomInit(float* data, int SIZE, int sparse);
void generalHepta(float* data, int j, int h, int i, int k);
void generalHeptaBdia(float* data, int j, int h, int i, int k);
void fillRow(float *data, int m, int j, int h, int i, int k);
void fillRowBdia(float *data, int m, int j, int h, int i, int k);
void fillBlock(float* data, int start, int blockSize, int mxSize, int value);


// for spMV
#define MXSIZE 128
#define BLOCKSIZE 8


/* Delete all entities */
static int publisher_shutdown(
    DDS_DomainParticipant *participant)
{
    DDS_ReturnCode_t retcode;
    int status = 0;

    if (participant != NULL) {
        retcode = DDS_DomainParticipant_delete_contained_entities(participant);
        if (retcode != DDS_RETCODE_OK) {
            printf("delete_contained_entities error %d\n", retcode);
            status = -1;
        }

        retcode = DDS_DomainParticipantFactory_delete_participant(
            DDS_TheParticipantFactory, participant);
        if (retcode != DDS_RETCODE_OK) {
            printf("delete_participant error %d\n", retcode);
            status = -1;
        }
    }

    /* RTI Data Distribution Service provides finalize_instance() method on
    domain participant factory for people who want to release memory used
    by the participant factory. Uncomment the following block of code for
    clean destruction of the singleton. */
    /*
    retcode = DDS_DomainParticipantFactory_finalize_instance();
    if (retcode != DDS_RETCODE_OK) {
        printf("finalize_instance error %d\n", retcode);
        status = -1;
    }
    */

    return status;
}

static int publisher_main(int domainId, int sample_count)
{
    // CUDA CODE TO BUILD MATRIX, VECTOR, AND OFFSETS
        const int SIZE = MXSIZE;

        int j, h, i, k, r, s, WS;
        float *X_h, *V_h, *Vc_h, *Vbdia_h, *hepta_h, *bdia_h;
        int *offsets;
        j = 4;
        h = 8;
        i = 8;
        k = 4;
        WS = 256;

        int mtxSize = i*h*j*k;
        int mtxBdiaSize = 7*k;

        // allocate array on host
        size_t vectSize = mtxSize*sizeof(float);
        size_t heptaSize = mtxSize*mtxSize*sizeof(float);
        size_t bdiaSize = mtxSize*mtxBdiaSize*sizeof(float);
        size_t offsetSize = mtxBdiaSize*sizeof(int);

        X_h= (float *) malloc(vectSize);
        V_h = (float *) malloc(vectSize);
        Vc_h = (float *) malloc(vectSize);
        Vbdia_h = (float *) malloc(vectSize);
        // hepta_h = (float *) malloc(heptaSize);
        bdia_h = (float *) malloc(bdiaSize);
        offsets = (int *) malloc(offsetSize);

        memset(X_h, 0, vectSize);
        memset(V_h, 0, vectSize);
        memset(Vc_h, 0, vectSize);
        memset(Vbdia_h, 0, vectSize);
        // memset(hepta_h, 0, heptaSize);
        memset(bdia_h, 0, bdiaSize);
        memset(offsets, 0, offsetSize);

        // initialize the offset for BDIA_GH
        for(r = 0; r < 7; r++ ){

            for(s = 0; s < k; s++){

              if (r == 0){

                 offsets[r*k+s] = s-(k*j*h);

              }else if (r == 1){

                offsets[r*k+s] = s-(k*j);

              }else if (r == 2)
              {
                offsets[r*k+s] = s-k;

              }else if (r == 3){

                offsets[r*k+s] = s;

              }else if (r == 4){

                offsets[r*k+s] = s+k;

              }else if (r == 5)
              {
                offsets[r*k+s] = s+(k*j);

              }else if (r == 6){
              
                offsets[r*k+s] = s+(k*j*h);
              }
            } 
        }

        // // initialize the X
        randomInit(X_h, mtxSize, 0);

        // // initialize the General Hepta on dense format
        // generalHepta(hepta_h, j, h, i, k);

        // // initialize the general hepta on bdia format
        generalHeptaBdia(bdia_h, j, h, i, k);
  // ############################PRINTING######################
        FILE *f = fopen("bdia.txt", "w");
        if (f == NULL)
        {
          printf("Error opening file!\n");
          exit(1);
        }


        fprintf(f,"####BDIA####\n");
        for(i=0;i<mtxSize;i++)
        {
          for(j=0;j<mtxBdiaSize;j++){
            // printf("%1.0f ", bdia_h[i*mtxBdiaSize+j]);
            fprintf(f,"%1.0f ", bdia_h[i+j*mtxSize]);
          }
          // printf("\n");
          fprintf(f,"\n" );
        }

        fclose(f);
        // Cleanup
        free(X_h);
        free(V_h);
        free(Vc_h);
        free(Vbdia_h);
        // free(hepta_h);
        free(bdia_h);
        free(offsets);
        // END OF CUDA CODE

    DDS_DomainParticipant *participant = NULL;
    DDS_Publisher *publisher = NULL;
    DDS_Topic *topic = NULL;
    DDS_DataWriter *writer = NULL;
    demo_rip_loadBalance_vectorDataWriter *demo_rip_loadBalance_vector_writer = NULL;
    demo_rip_loadBalance_vector *instance = NULL;
    DDS_ReturnCode_t retcode;
    DDS_InstanceHandle_t instance_handle = DDS_HANDLE_NIL;
    const char *type_name = NULL;
    int count = 0;  
    struct DDS_Duration_t send_period = {4,0};

    /* To customize participant QoS, use 
    the configuration file USER_QOS_PROFILES.xml */
    participant = DDS_DomainParticipantFactory_create_participant(
        DDS_TheParticipantFactory, domainId, &DDS_PARTICIPANT_QOS_DEFAULT,
        NULL /* listener */, DDS_STATUS_MASK_NONE);
    if (participant == NULL) {
        printf("create_participant error\n");
        publisher_shutdown(participant);
        return -1;
    }

    /* To customize publisher QoS, use 
    the configuration file USER_QOS_PROFILES.xml */
    publisher = DDS_DomainParticipant_create_publisher(
        participant, &DDS_PUBLISHER_QOS_DEFAULT, NULL /* listener */,
        DDS_STATUS_MASK_NONE);
    if (publisher == NULL) {
        printf("create_publisher error\n");
        publisher_shutdown(participant);
        return -1;
    }

    /* Register type before creating topic */
    type_name = demo_rip_loadBalance_vectorTypeSupport_get_type_name();
    retcode = demo_rip_loadBalance_vectorTypeSupport_register_type(
        participant, type_name);
    if (retcode != DDS_RETCODE_OK) {
        printf("register_type error %d\n", retcode);
        publisher_shutdown(participant);
        return -1;
    }

    /* To customize topic QoS, use 
    the configuration file USER_QOS_PROFILES.xml */
    topic = DDS_DomainParticipant_create_topic(
        participant, "Example demo_rip_loadBalance_vector",
        type_name, &DDS_TOPIC_QOS_DEFAULT, NULL /* listener */,
        DDS_STATUS_MASK_NONE);
    if (topic == NULL) {
        printf("create_topic error\n");
        publisher_shutdown(participant);
        return -1;
    }

    /* To customize data writer QoS, use 
    the configuration file USER_QOS_PROFILES.xml */
    writer = DDS_Publisher_create_datawriter(
        publisher, topic,
        &DDS_DATAWRITER_QOS_DEFAULT, NULL /* listener */, DDS_STATUS_MASK_NONE);
    if (writer == NULL) {
        printf("create_datawriter error\n");
        publisher_shutdown(participant);
        return -1;
    }
    demo_rip_loadBalance_vector_writer = demo_rip_loadBalance_vectorDataWriter_narrow(writer);
    if (demo_rip_loadBalance_vector_writer == NULL) {
        printf("DataWriter narrow error\n");
        publisher_shutdown(participant);
        return -1;
    }

    /* Create data sample for writing */
    instance = demo_rip_loadBalance_vectorTypeSupport_create_data_ex(DDS_BOOLEAN_TRUE);
    if (instance == NULL) {
        printf("demo_rip_loadBalance_vectorTypeSupport_create_data error\n");
        publisher_shutdown(participant);
        return -1;
    }

    /* For a data type that has a key, if the same instance is going to be
    written multiple times, initialize the key here
    and register the keyed instance prior to writing */
    /*
    instance_handle = demo_rip_loadBalance_vectorDataWriter_register_instance(
        demo_rip_loadBalance_vector_writer, instance);
    */

    /* Main loop */
    for (count=0; (sample_count == 0) || (count < sample_count); ++count) {

        printf("Writing demo_rip_loadBalance_vector, count %d\n", count);

        /* Modify the data to be written here */

        /* Write data */
        retcode = demo_rip_loadBalance_vectorDataWriter_write(
            demo_rip_loadBalance_vector_writer, instance, &instance_handle);
        if (retcode != DDS_RETCODE_OK) {
            printf("write error %d\n", retcode);
        }

        NDDS_Utility_sleep(&send_period);
    }

    /*
    retcode = demo_rip_loadBalance_vectorDataWriter_unregister_instance(
        demo_rip_loadBalance_vector_writer, instance, &instance_handle);
    if (retcode != DDS_RETCODE_OK) {
        printf("unregister instance error %d\n", retcode);
    }
    */

    /* Delete data sample */
    retcode = demo_rip_loadBalance_vectorTypeSupport_delete_data_ex(instance, DDS_BOOLEAN_TRUE);
    if (retcode != DDS_RETCODE_OK) {
        printf("demo_rip_loadBalance_vectorTypeSupport_delete_data error %d\n", retcode);
    }

    /* Cleanup and delete delete all entities */         
    return publisher_shutdown(participant);
}

int main(int argc, char *argv[])
{
    int domainId = 0;
    int sample_count = 0; /* infinite loop */

    if (argc >= 2) {
        domainId = atoi(argv[1]);
    }
    if (argc >= 3) {
        sample_count = atoi(argv[2]);
    }

    /* Uncomment this to turn on additional logging
    NDDS_Config_Logger_set_verbosity_by_category(
        NDDS_Config_Logger_get_instance(),
        NDDS_CONFIG_LOG_CATEGORY_API, 
        NDDS_CONFIG_LOG_VERBOSITY_STATUS_ALL);
    */

    return publisher_main(domainId, sample_count);
}

//###################INITIALIZATION HEPTA###########################
void generalHepta(float* data, int j, int h, int i, int k)
{
  int mj, mh, mi;
  
  for (mi=0; mi<i; mi++) {
    for (mh=0; mh<h; mh++) {
      for (mj=0; mj<j; mj++) {
        int m = mj+ mh*j + mi*h*j;
        fillRow(data, m, j, h, i, k);
      }
    }
  }
}

void fillRow(float *data, int m, int j, int h, int i, int k){
  int mxSize = j*h*i*k;
  int blockSize = k;
  int mainStart, start;
  // int maxIdx = mxSize*mxSize-1;
  
  //main diagonal
  mainStart = m*mxSize*k+m*k;
  fillBlock(data, mainStart, blockSize, mxSize, 4);
  
  int min = m*k*mxSize;
  int max = min+mxSize;
  
  //diagonal 1
  start = mainStart-(h*j*k);
  if (min<=start) {
    fillBlock(data, start, blockSize, mxSize, 1);
  }
  
  //diagonal 2
  start = mainStart-j*k;
  if (min<=start) {
    fillBlock(data, start, blockSize, mxSize, 2);
  }
  
  //diagonal 3
  start = mainStart-1*k;
  if (min<=start) {
    fillBlock(data, start, blockSize, mxSize, 3);
  }

  //diagonal 5
  start = mainStart+1*k;
  if (start<max) {
    fillBlock(data, start, blockSize, mxSize, 5);
  }

  //diagonal 6
  start = mainStart+j*k;
  if (start<max) {
    fillBlock(data, start, blockSize, mxSize, 6);
  }
  
  //diagonal 7
  start = mainStart+(h*j*k);
  if (start<max) {
    fillBlock(data, start, blockSize, mxSize, 7);
  }

}

void generalHeptaBdia(float* data, int j, int h, int i, int k)
{

  // printf("generalHeptaBdia\n");
  int mj, mh, mi;
  
  for (mi=0; mi<i; mi++) {
    for (mh=0; mh<h; mh++) {
      for (mj=0; mj<j; mj++) {
        int m = mj+ mh*j + mi*h*j;
        fillRowBdia(data, m, j, h, i, k);
      }
    }
  }
}

void fillRowBdia(float *data, int m, int j, int h, int i, int k){
  // printf("fillRowBdia\n");
  int mxSize = j*h*i*k;
  // int rowSize = j*h*i*k;
  int colSize = 7*k;
  int blockSize = k;
  int mainStart, mainStartBdia, start, startBdia;
  // int maxIdx = rowSize*colSize-1;
  
  //main diagonal
  mainStart = m*mxSize*k+m*k; // main diagonal start of hepta
  mainStartBdia = m*k+mxSize*3*k; // main diagonal start of bdia
//  printf("m = %d, k=%d, mainStartBdia = %d\n", m,k,mainStartBdia);
  fillBlock(data, mainStartBdia, blockSize, mxSize, 4);
  
  int min = m*k*mxSize;
  int max = min+mxSize;
  
  //diagonal 1
  start = mainStart-(h*j*k);
  startBdia = mainStartBdia-(3*k*mxSize);
  if (min<=start) {
    fillBlock(data, startBdia, blockSize, mxSize, 1);
  }
  
  //diagonal 2
  start = mainStart-j*k;
  startBdia = mainStartBdia-(2*k*mxSize);
  if (min<=start) {
    fillBlock(data, startBdia, blockSize, mxSize, 2);
  }
  
  //diagonal 3
  start = mainStart-1*k;
  startBdia = mainStartBdia-(1*k*mxSize);
  if (min<=start) {
    fillBlock(data, startBdia, blockSize, mxSize, 3);
  }

  //diagonal 5
  start = mainStart+1*k;
  startBdia = mainStartBdia+(1*k*mxSize);
  if (start<max) {
    fillBlock(data, startBdia, blockSize, mxSize, 5);
  }
  
  //diagonal 6
  start = mainStart+j*k;
  startBdia = mainStartBdia+(2*k*mxSize);
  if (start<max) {
    fillBlock(data, startBdia, blockSize, mxSize, 6);
  }
  
  //diagonal 7
  start = mainStart+(h*j*k);
  startBdia = mainStartBdia+(3*k*mxSize);
  if (start<max) {
    fillBlock(data, startBdia, blockSize, mxSize, 7);
  }
}

void fillBlock(float* data, int start, int blockSize, int mxSize, int value)
{
  // printf("fillBlock\n");
  int i,j;
  for (j=0; j<blockSize; j++) {
    for (i=0; i<blockSize; i++) {
      data[start+i+j*mxSize] = value;
    }
  }
}

// fillBlockDown is the transpose version of block up so the matrix will be symmetric
void fillBlockDown(float* data, int start, int blockSize, int mxSize, float value)
{
  int i,j;
  for (i=0; i<blockSize; i++) {
    for (j=0; j<blockSize; j++) {
      data[start+i+j*mxSize] = value;
    }
  }
}

void fillBlockSym(float* data, int start, int blockSize, int mxSize, int *value)
{
  int i,j,k;
  k=0;
  for (i=0; i<blockSize; i++) {
    for (j=0; j<blockSize; j++) {
      data[start+i+j*mxSize] = value[k];
      k++;
    }
  }
}

// fillBlockDown is the transpose version of block up so the matrix will be symmetric
void fillBlockSymDown(float* data, int start, int blockSize, int mxSize, float *value)
{
  int i,j,k;
  k=0;
  for (j=0; j<blockSize; j++) {
    for (i=0; i<blockSize; i++) {
      data[start+i+j*mxSize] = value[k];
      k++;
    }
  }
}


//void fill_half_block(float* data, int k, int blockSize){
//  int i,j;
//  for (i=0; i<BLOCKSIZE/2; i++) {
//    for (j=0; j<BLOCKSIZE; j++) {
//      data[i+k+j*MXSIZE] = 1;
//    }
//  }
//}

void randomInit(float* data, int SIZE,int sparse)
{
    float r;
    int row;
    int col;
    int i;
    float val_r = 1;
    //        float val_r = (float)rand() / (float)RAND_MAX;
    
    if(sparse == 1){
        //           int nze = size * (float)(1.0-12);
        //           for (int i = 0; i < nze; ++i){
        //                r = 12456789;
        //                data[i] = r;
        //           }
        //           std::random_shuffle(data, data+size);
    }
    else if(sparse == 2){ //hepta diagonal
        int mnz = 22;
        float pattern[22];
        r = val_r;
        pattern[0] = r;
        r = val_r;
        pattern[1] = r;
        r = val_r;
        pattern[2] = 0;
        pattern[3] = 0;
        pattern[4] = 0;
        pattern[5] = r;
        r = val_r;
        pattern[6] = r;
        r = val_r;
        pattern[7] = 0;
        pattern[8] = r;
        r = val_r;
        pattern[9] = r;
        r = val_r;
        pattern[10] = r;
        r = val_r;
        pattern[11] = r;
        r = val_r;
        pattern[12] = r;
        r = val_r;
        pattern[13] = r;
        r = val_r;
        pattern[14] = 0;
        pattern[15] = r;
        r = val_r;
        pattern[16] = r;
        r = val_r;
        pattern[17] = 0;
        pattern[18] = 0;
        pattern[19] = 0;
        pattern[20] = r;
        r = val_r;
        pattern[21] = r;
        
        for(row=10; row<SIZE; row+=2){
            int count=0;
            for(col=(row-10); col<SIZE && count < mnz; col++){
                data[row*SIZE+col] = pattern[count];
                data[(row+1)*SIZE+col] = pattern[count++];
            }
        }
        for(row=9; row >= 0; row-=2){
            int count = 10-row+1;
            for(col=0; col<SIZE && count<mnz; col++){
                data[row*SIZE+col] = pattern[count];
                data[(row-1)*SIZE+col] = pattern[count++];
            }
        }
    }
    else{
        for (i = 0; i < SIZE; ++i){
            r = val_r;
            data[i] = r;
        }
    }
    
}